import psutil, os, stat, sys, time, traceback
from tempfile import NamedTemporaryFile
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from .streaming_log import StreamingLog
from shortuuid import ShortUUID
import atexit
from threading import Thread
from .config import Config
import logging, socket
from wandb import __stage_dir__, Error
logger = logging.getLogger(__name__)

class Echo(object):
    def __init__(self, log):
        self.terminal = sys.stdout
        self.log = log

    def write(self, message):
        self.terminal.write(message)
        #TODO: ThreadPool
        self.thread = Thread(target=self.log.write, args=(message,))
        self.thread.start()

    def flush(self):
        self.terminal.flush()

    def close(self, failed=False):
        self.thread.join()
        if failed:
            #TODO: unfortunate line_buffer access
            self.log.push([[self.log.line_buffer.line_number + 1, "ERROR: %s\n" % failed, "error"]])
            sys.stderr.write("ERROR: %s\n" % failed)
            failed = True
        self.log.heartbeat(complete=True, failed=failed)

class ExitHooks(object):
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        self._orig_excepthook = sys.excepthook
        sys.exit = self.exit
        sys.excepthook = self.excepthook

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def excepthook(self, exc_type, exc, *args):
        self.exception = exc
        self._orig_excepthook(exc_type, exc, *args)

class Sync(object):
    """Watches for files to change and automatically pushes them
    """
    def __init__(self, api, run=None, project=None, tags=[], datasets=[], config={}, description=None):
        #1.6 million 6 character combinations
        runGen = ShortUUID(alphabet=list("0123456789abcdefghijklmnopqrstuvwxyz"))
        self.run = run or runGen.random(6)
        self._project = project or api.config("project")
        self._entity = api.config("entity")
        logger.debug("Initialized sync for %s/%s", self._project, self.run)
        self._dpath = ".wandb/description.md"
        self._description = description or os.path.exists(self._dpath) and open(self._dpath).read()
        self.config = Config(config)
        self._proc = psutil.Process(os.getpid())
        self._api = api
        self._tags = tags
        self._handler = PatternMatchingEventHandler()
        self._handler.on_created = self.add
        self._handler.on_modified = self.push
        self.url = "{base}/{entity}/{project}/runs/{run}".format(
            project=self._project,
            entity=self._entity,
            run=self.run,
            base="https://app.wandb.ai"
        )
        self.log = StreamingLog(self.run)
        self._hooks = ExitHooks()
        self._hooks.hook()
        self._observer = Observer()
        self._watch_dir = os.getcwd()
        self._observer.schedule(self._handler, self._watch_dir, recursive=True)

    def watch(self, files):
        try:
            #TODO: better failure handling
            self._api.upsert_bucket(name=self.run, project=self._project, entity=self._entity, 
                config=self.config.__dict__, description=self._description, host=socket.gethostname())
            self._handler._patterns = [
                os.path.join(self._watch_dir, os.path.normpath(f)) for f in files]
            # temporary until we switch to sending all files within a dedicated run
            # directory.
            self._handler._patterns += ['*wandb-summary.json', '*wandb-history.csv']
            # Ignore hidden files/folders
            self._handler._ignore_patterns = ['*/.*']
            if os.path.exists(__stage_dir__+"diff.patch"):
                self._api.push("{project}/{run}".format(
                    project=self._project,
                    run=self.run
                ), {"diff.patch": open(__stage_dir__+"diff.patch", "rb")})
            self._observer.start()
            print("Syncing %s" % self.url)
            # Piped mode
            if self.source_proc:
                self.log.write(" ".join(self.source_proc.cmdline())+"\n\n")
                line = sys.stdin.readline()
                while line:
                    sys.stdout.write(line)
                    self.log.write(line)
                    #TODO: push log every few minutes...
                    line = sys.stdin.readline()
                self.stop()
            else:
                self.log.write(" ".join(psutil.Process(os.getpid()).cmdline())+"\n\n")
                # let's hijack stdout
                sys.stdout = Echo(self.log)
                logger.debug("Swapped stdout")
                #TODO: stderr
                atexit.register(self.stop)
        except KeyboardInterrupt:
            self.stop()
        except Error:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            print("!!! Fatal W&B Error: %s" % exc_value)
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            logger.error('\n'.join(lines))

    def stop(self):
        # This is a a heuristic delay to catch files that were written just before
        # the end of the script. This is unverified, but theoretically the file
        # change notification process used by watchdog (maybe inotify?) is
        # asynchronous. It's possible we could miss files if 10s isn't long enough.
        # TODO: Guarantee that all files will be saved.
        print("Script ended, waiting for final file modifications.")
        time.sleep(10.0)
        self.log.tempfile.flush()
        print("Pushing log")
        slug = "{project}/{run}".format(
            project=self._project,
            run=self.run
        )
        self._api.push(slug, {"training.log": open(self.log.tempfile.name, "rb")})
        os.path.exists(self._dpath) and os.remove(self._dpath)
        print("Synced %s" % self.url)
        sys.stdout.close(failed=self._hooks.exception)
        self.log.close()
        try:
            self._observer.stop()
            self._observer.join()
        #TODO: py2 TypeError: PyCObject_AsVoidPtr called with null pointer
        except TypeError:
            pass
        #TODO: py3 SystemError: <built-in function stop> returned a result with an error set
        except SystemError:
            pass

    #TODO: limit / throttle the number of adds / pushes
    def add(self, event):
        self.push(event)

    #TODO: is this blocking the main thread?
    def push(self, event):
        if os.stat(event.src_path).st_size == 0 or os.path.isdir(event.src_path):
            return None
        fileName = os.path.relpath(event.src_path, self._watch_dir)
        if logger.parent.handlers[0]:
            debugLog = logger.parent.handlers[0].stream
        else:
            debugLog = None
        self._api.push(self._project, [fileName], bucket=self.run, 
                       description=self._description, progress=debugLog)

    @property
    def source_proc(self):
        mode = os.fstat(0).st_mode
        if not stat.S_ISFIFO(mode):
            # stdin is not a pipe
            return None
        else:
            source = self._proc.parent().children()[0]
            return None if source == self._proc else source

    def echo(self):
        print(sys.stdin.read())



import json
import numpy as np

class WandBJSONEncoder(json.JSONEncoder):
    """A JSON Encoder that handles some extra types."""
    def default(self, obj):
        # TODO: Some of this is just guessing. Be smarter.
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return np.asscalar(obj)
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        return json.JSONEncoder.default(self, obj)

def json_dumps_safer(obj, **kwargs):
    """Convert obj to json, with some extra encodable types."""
    return json.dumps(obj, cls=WandBJSONEncoder, **kwargs)

def make_json_if_not_number(v):
    """If v is not a basic type convert it to json."""
    if isinstance(v, (float, int)):
        return v
    return util.json_dumps_safer(v)


import io
import tempfile
import time
import sys
from requests import Session
from wandb import Api
import threading
import traceback
import logging
import signal
import os
import logging
import six
logger = logging.getLogger(__name__)

STALE_SECONDS = 2
RATE_LIMIT = 1
MAX_LINES = 5
HTTP_TIMEOUT = 3
HEARTBEAT_INTERVAL = 15

class LineBuffer(io.FileIO):
    """Writes to a tempfile while buffering lines and deduping repeated
    lines with a carriage return.  Calls push every MAX_LINES if under RATE_LIMIT,
    or atleast every STALE_SECONDS after the last line was written.
    """
    def __init__(self, path, push = lambda *args: None):
        super(LineBuffer, self).__init__(path, "w")
        self.buf = bytearray()
        self.lock = threading.RLock()
        self.line_number = 0
        self.lines = []
        self.posted = time.time()
        self.push = push
        self.posting = False

    @property
    def stale(self):
        return (not self.posting and
                len(self.lines) > 0 and
                (time.time() - self.posted) >= STALE_SECONDS)

    @property
    def rate_limit(self):
        return time.time() - self.posted < RATE_LIMIT

    def flush(self):
        """This assumes flush is called once per line (\n, \r, or \r\n),
        io.TextIOWrapper ensures this is true"""
        super(LineBuffer, self).flush()
        if len(self.buf) == 0:
            return False
        with self.lock:
            self.line_number += 1
            lines = len(self.lines) + 1
            if(lines > 1 and self.lines[-1][1].endswith("\r")):
                self.line_number -= 1
                self.lines.pop()
            self.lines.append([self.line_number, self.buf.decode("utf-8")])
            del self.buf[:]
            if (lines >= MAX_LINES and not self.rate_limit) or self.stale:
                logger.debug("Pushing %i lines (stale? %s)", lines, self.stale)
                to_push = self.lines[:]
                if to_push[-1][1].endswith("\r"):
                    del self.lines[:-1]
                else:
                    del self.lines[:]
                self.posting = True
                #Release the lock to allow multiple concurrent pushes
                self.lock.release()
                if self.push(to_push):
                    self.lock.acquire()
                    self.posted = time.time()
                else:
                    self.lock.acquire()
                    self.lines = to_push + self.lines
                self.posting = False

    def write(self, chars):
        super(LineBuffer, self).write(chars)
        with self.lock:
            self.buf.extend(chars)

class StreamingLog(io.TextIOWrapper):
    """Manages the WandB client and tempfile for log storage"""
    api = Api()
    config = api.config()
    endpoint = "{base}/{entity}/{project}/%s/logs".format(
        base=config['base_url'],
        entity=config.get("entity"),
        project=config.get("project")
    )

    def __init__(self, run, level=logging.INFO):
        self.run = run
        self.client = Session()
        self.client.auth = ("api", StreamingLog.api.api_key)
        self.client.timeout = HTTP_TIMEOUT
        self.client.headers.update({
            'User-Agent': StreamingLog.api.user_agent,
        })
        self.tempfile = tempfile.NamedTemporaryFile(mode='wb')
        self.pushed_at = time.time()
        self.pushing = 0
        self.line_buffer = LineBuffer(self.tempfile.name, push=self.push)
        #Schedule heartbeat TODO: unix only, use threads
        signal.signal(signal.SIGALRM, self.heartbeat)
        signal.setitimer(signal.ITIMER_REAL, HEARTBEAT_INTERVAL / 2)
        super(StreamingLog, self).__init__(self.line_buffer, line_buffering=True, 
                                           newline='')
    
    def heartbeat(self, *args, **kwargs):
        complete = kwargs.get("complete")
        failed = kwargs.get("failed")
        if complete or time.time() - self.pushed_at > HEARTBEAT_INTERVAL:
            logger.debug("Heartbeat sent at %s", time.time())
            self.client.post(StreamingLog.endpoint % self.run, json={'complete': complete, 'failed': failed})

    def write(self, chars):
        if six.PY2:
            chars = unicode(chars)
        super(StreamingLog, self).write(chars)

    def push(self, lines):
        try:
            self.pushing += 1
            res = self.client.post(StreamingLog.endpoint % self.run, 
                json={'lines': lines})
            res.raise_for_status()
            self.pushing -= 1
            self.pushed_at = time.time()
            return res
        except Exception as err:
            #TODO: Is this the right thing to do?
            if self.pushing > 3:
                raise err
            logger.error('Unable to post to WandB: %s', err)
            self.pushing -= 1
            return False


import csv
import os

from wandb import util

HISTORY_FNAME = 'wandb-history.csv'

class History(object):
    """Used to store data that changes over time during runs.

    This file can be read back into Python by converting all calling json.load()
    on all string values.
    """
    def __init__(self, field_names, out_dir='.'):
        self.out_file = open(os.path.join(out_dir, HISTORY_FNAME), 'w')
        self.out_csv = csv.DictWriter(self.out_file, fieldnames=field_names)
        self.out_csv.writeheader()
        self.out_file.flush()
    
    def add(self, row):
        row = {k: util.make_json_if_not_number(v) for k, v in row.items()}
        self.out_csv.writerow(row)
        self.out_file.flush()

    def close(self):
        self.out_file.close()


# -*- coding: utf-8 -*-

import click, sys
from wandb import Api, Error, Sync, Config, __version__, __stage_dir__
import random, time, os, re, netrc, logging, json, glob, io, stat
from functools import wraps
from click.utils import LazyFile
from click.exceptions import BadParameter, ClickException
import inquirer
import sys, traceback

logger = logging.getLogger(__name__)

def write_netrc(host, entity, key):
    """Add our host and key to .netrc"""
    try:
        normalized_host = host.split("/")[-1].split(":")[0]
        print("Appending to netrc %s" % os.path.expanduser('~/.netrc')) 
        with open(os.path.expanduser('~/.netrc'), 'a') as f:
            f.write("""machine {host}
        login {entity}
        password {key}
    """.format(host=normalized_host, entity=entity, key=key))
        os.chmod(os.path.expanduser('~/.netrc'), stat.S_IRUSR | stat.S_IWUSR)
    except IOError as e:
        click.secho("Unable to read ~/.netrc", fg="red")
        return None

def display_error(func):
    """Function decorator for catching common errors and re-raising as wandb.Error"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Error as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            logger.error('\n'.join(lines))
            raise ClickException(e)
            
    return wrapper

def editor(content='', marker='# Enter a description, markdown is allowed!\n'):
    message = click.edit(content + '\n\n' + marker)
    if message is not None:
        return message.split(marker, 1)[0].rstrip('\n')
        
api = Api()
#TODO: Is this the best way to do this?
CONTEXT=dict(default_map=api.config())

class BucketGroup(click.Group):
    @display_error
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        try:
            project, bucket = api.parse_slug(cmd_name)
        except Error:
            return None
        #TODO: This is hacky as hell
        description = None
        if '-m' in sys.argv:
            description = sys.argv[sys.argv.index('-m') + 1]
        elif '--description' in sys.argv:
            description = sys.argv[sys.argv.index('--description') + 1]
        sync = Sync(api, project=project, run=bucket, description=description)
        if sync.source_proc:
            files = sys.argv[2:]
            sync.watch(files)
            return click.Command("sync", context_settings={'allow_extra_args': True})
        else:
            return None

@click.command(cls=BucketGroup)
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx):
    """Weights & Biases

If no command is specified and input is piped, the source command and it's 
output will be saved to the bucket and the files uploaded when modified.

   ./train.sh arg1 arg2 | wandb imagenet/v2 model.json weights.h5
    """
    pass

@cli.command(context_settings=CONTEXT, help="List projects")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def projects(entity, display=True):
    projects = api.list_projects(entity=entity)
    if len(projects) == 0:
        message = "No projects found for %s" % entity
    else:
        message = 'Latest projects for "%s"' % entity
    if display:
        click.echo(click.style(message, bold=True))
        for project in projects:
            click.echo("".join(
                (click.style(project['name'], fg="blue", bold=True),
                " - ",
                str(project['description']).split("\n")[0])
            ))
    return projects

@cli.command(context_settings=CONTEXT, help="List buckets in a project")
@click.argument("project", envvar='WANDB_PROJECT')
@click.option("--project", "-p", prompt=True, envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def buckets(project, entity):
    click.echo(click.style('Latest buckets for project "%s"' % project, bold=True))
    buckets = api.list_buckets(project, entity=entity)
    for bucket in buckets:
        click.echo("".join(
            (click.style(bucket['name'], fg="blue", bold=True), 
            " - ", 
            (bucket['description'] or "").split("\n")[0])
        ))

@cli.command(context_settings=CONTEXT, help="List staged & remote files")
@click.argument("bucket", envvar='WANDB_BUCKET')
@click.option("--config/--no-config", help="Show the current configuration", default=False)
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@display_error
def status(bucket, config, project):
    if config:
        click.echo(click.style("Current Configuration", bold=True) + " (%s)" % api.config_file)
        config = api.config()
        click.echo(json.dumps(
            config,
            sort_keys=True,
            indent=2,
            separators=(',', ': ')
        ))
        click.echo(click.style("Logged in?", bold=True) + " %s\n" % bool(api.api_key))
    project, bucket = api.parse_slug(bucket, project=project)
    parser = api.config_parser
    parser.read(".wandb/config")
    if parser.has_option("default", "files"):
        existing = set(parser.get("default", "files").split(","))
    else:
        existing = set()
    remote = api.download_urls(project, bucket)
    not_synced = set()
    remote_names = set([name for name in remote])
    for file in existing:
        meta = remote.get(file)
        if meta and not api.file_current(file, meta['md5']):
            not_synced.add(file)
        elif not meta:
            not_synced.add(file)
    #TODO: remove items that exists and have the md5
    only_remote = remote_names.difference(existing)
    up_to_date = existing.difference(only_remote).difference(not_synced)
    click.echo('File status for '+ click.style('"%s/%s" ' % (project, bucket), bold=True))
    if len(not_synced) > 0:
        click.echo(click.style('Push needed: ', bold=True) + click.style(", ".join(not_synced), fg="red"))
    if len(only_remote) > 0:
        click.echo(click.style('Pull needed: ', bold=True) + click.style(", ".join(only_remote), fg="red"))
    if len(up_to_date) > 0:
        click.echo(click.style('Up to date: ', bold=True) + click.style(", ".join(up_to_date), fg="green"))
    if len(existing) == 0:
        click.echo(click.style("No files configured, add files with `wandb add filename`", fg="red"))

@cli.command(context_settings=CONTEXT, help="Store notes for a future training run")
@display_error
def describe():
    path = __stage_dir__+'description.md'
    existing = (os.path.exists(path) and open(path).read()) or ''
    description = editor(existing)
    if description:
        with open(path, 'w') as file:
            file.write(description)
    click.echo("Notes stored for next training run\nCalling wandb.sync() in your training script will persist them.")

@cli.command(context_settings=CONTEXT, help="Restore code and config state for bucket")
@click.argument("bucket", envvar='WANDB_BUCKET')
@click.option("--branch/--no-branch", default=True, help="Whether to create a branch or checkout detached")
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def restore(bucket, branch, project, entity):
    project, bucket = api.parse_slug(bucket, project=project)
    commit, json_config, patch = api.bucket_config(project, bucket=bucket, entity=entity)
    if commit:
        branch_name = "wandb/%s" % bucket
        if branch and branch_name not in api.git.repo.branches:
            api.git.repo.git.checkout(commit, b=branch_name)
            click.echo("Created branch %s" % click.style(branch_name, bold=True))
        elif branch:
            click.secho("Using existing branch, run `git branch -D %s` from master for a clean checkout" % branch_name, fg="red")
            api.git.repo.git.checkout(branch_name)
        else:
            click.secho("Checking out %s in detached mode" % commit)
            api.git.repo.git.checkout(commit)
    
    if patch:
        with open(__stage_dir__+"diff.patch", "w") as f:
            f.write(patch)
        api.git.repo.git.execute(['git', 'apply', __stage_dir__+'diff.patch'])
        click.echo("Applied patch")

    config = Config()
    config.load_json(json_config)
    config.persist()
    click.echo("Restored config variables")

@cli.command(context_settings=CONTEXT, help="Add staged files")
@click.argument("files", type=click.File('rb'), nargs=-1)
@click.option("--project", "-p", prompt=True, envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@display_error
def add(files, project):
    parser = api.config_parser
    parser.read(".wandb/config")
    if not parser.has_section("default"):
        raise ClickException("Directory not configured, run `wandb init` before adding files.")
    if parser.has_option("default", "files"):
        existing = parser.get("default", "files").split(",")
    else:
        existing = []
    stagedFiles = set(existing + [file.name for file in files])
    parser.set("default", "files", ",".join(stagedFiles))
    with open('.wandb/config', 'w') as configfile:
        parser.write(configfile)
    click.echo(click.style('Staged files for "%s": ' % project, bold=True) + ", ".join(stagedFiles))

@cli.command(context_settings=CONTEXT, help="Remove staged files")
@click.argument("files", nargs=-1)
@click.option("--project", "-p", prompt=True, envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@display_error
def rm(files, project):
    parser = api.config_parser
    parser.read(".wandb/config")
    if parser.has_option("default", "files"):
        existing = parser.get("default", "files").split(",")
    else:
        existing = []
    for file in files:
        if file not in existing:
            raise ClickException("%s is not staged" % file)
        existing.remove(file)
    parser.set("default", "files", ",".join(existing))
    with open('.wandb/config', 'w') as configfile:
        parser.write(configfile)
    click.echo(click.style('Staged files for "%s": ' % project, bold=True) + ", ".join(existing))
    
@cli.command(context_settings=CONTEXT, help="Push files to Weights & Biases")
@click.argument("bucket", envvar='WANDB_BUCKET')
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@click.option("--description", "-m", help="A description to associate with this upload.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@click.option("--force/--no-force", "-f", default=False, help="Whether to force git tag creation.")
@click.argument("files", type=click.File('rb'), nargs=-1)
@click.pass_context
@display_error
def push(ctx, bucket, project, description, entity, force, files):
    #TODO: do we support the case of a bucket with the same name as a file?
    if os.path.exists(bucket):
        raise BadParameter("Bucket is required if files are specified.")
    project, bucket = api.parse_slug(bucket, project=project)

    click.echo("Updating project: {project}/{bucket}".format(
        project=click.style(project, bold=True), bucket=bucket))
    if description is None:
        description = editor()
    
    candidates = []
    if len(files) == 0:
        if api.config().get("files"):
            fileNames = api.config()['files'].split(",")
            files = [LazyFile(fileName, 'rb') for fileName in fileNames]
        else:
            patterns = ("*.h5", "*.hdf5", "*.json", "*.meta", "*checkpoint*")
            for pattern in patterns:
                candidates.extend(glob.glob(pattern))          
            if len(candidates) == 0:
                raise BadParameter("Couldn't auto-detect files, specify manually or use `wandb.add`", param_hint="FILES")

            choices = inquirer.prompt([inquirer.Checkbox('files', message="Which files do you want to push? (left and right arrows to select)", 
                choices=[c for c in candidates])])
            files = [LazyFile(choice, 'rb') for choice in choices['files']]

    if len(files) > 10:
        raise BadParameter("A maximum of 10 files can be in a single bucket.", param_hint="FILES")
    #TODO: Deal with files in a sub directory
    api.push(project, files=[f.name for f in files], bucket=bucket, 
        description=description, entity=entity, force=force, progress=sys.stdout)

@cli.command(context_settings=CONTEXT, help="Pull files from Weights & Biases")
@click.argument("bucket", envvar='WANDB_BUCKET')
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you want to download.")
@click.option("--kind", "-k", default="all", type=click.Choice(['all', 'model', 'weights', 'other']))
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def pull(project, bucket, kind, entity):
    project, bucket = api.parse_slug(bucket, project=project)

    urls = api.download_urls(project, bucket=bucket, entity=entity)
    if len(urls) == 0:
        raise ClickException("Bucket is empty")
    click.echo("Downloading: {project}/{bucket}".format(
        project=click.style(project, bold=True), bucket=bucket
    ))

    for name in urls:
        if api.file_current(name, urls[name]['md5']):
            click.echo("File %s is up to date" % name)
        else:
            length, response = api.download_file(urls[name]['url'])
            with click.progressbar(length=length, label='File %s' % name,
                                fill_char=click.style('&', fg='green')) as bar:
                with open(name, "wb") as f:
                    for data in response.iter_content(chunk_size=4096):
                        f.write(data)
                        bar.update(len(data))

@cli.command(context_settings=CONTEXT, help="Login to Weights & Biases")
@display_error
def login():
    # Import in here for performance reasons
    import webbrowser
    #TODO: use Oauth and a local webserver: https://community.auth0.com/questions/6501/authenticating-an-installed-cli-with-oidc-and-a-th
    url = "https://app.wandb.ai/profile"
    #TODO: google cloud SDK check_browser.py
    launched = webbrowser.open_new_tab(url)
    if launched:
        click.echo('Opening [{0}] in a new tab in your default browser.'.format(url))
    else:
        click.echo("You can find your API keys here: {0}".format(url))
    key = click.prompt("{warning} Paste an API key from your profile".format(
            warning=click.style("Not authenticated!", fg="red")), default="")
    host = api.config()['base_url']
    if key:
        #TODO: get the username here...
        #username = api.viewer().get('entity', 'models')
        write_netrc(host, "user", key)

@cli.command(context_settings=CONTEXT, help="Configure a directory with Weights & Biases")
@click.pass_context
@display_error
def init(ctx):
    if(os.path.exists(".wandb")):
        click.confirm(click.style("This directory is already configured, should we overwrite it?", fg="red"), abort=True)
    click.echo(click.style("Let's setup this directory for W&B!", fg="green", bold=True))
    
    if api.api_key is None:
        ctx.invoke(login)

    entity = click.prompt("What username or org should we use?", default=api.viewer().get('entity', 'models'))
    #TODO: handle the case of a missing entity
    result = ctx.invoke(projects, entity=entity, display=False)

    if len(result) == 0:
        project = click.prompt("Enter a name for your first project")
        description = editor()
        api.create_project(project, entity=entity, description=description)
    else:
        project_names = [project["name"] for project in result]
        question = inquirer.List('project', message="Which project should we use?", choices=project_names + ["Create New"])
        project = inquirer.prompt([question])['project']
        #TODO: check with the server if the project exists
        if project == "Create New":
            project = click.prompt("Enter a name for your new project")
            description = editor()
            api.create_project(project, entity=entity, description=description)

    ctx.invoke(config_init, False)

    with open(".wandb/config", "w") as file:
        file.write("[default]\nentity: {entity}\nproject: {project}".format(entity=entity, project=project))

    with open(".wandb/.gitignore", "w") as file:
        file.write("*\n!config")

    click.echo(click.style("This directory is configured!  Try these next:\n", fg="green")+ 
        """
* Track runs by calling sync in your training script `{flags}`.
* Run `{push}` to manually add a file.
* `{config}` to add or change configuration defaults.
* Pull popular models into your project with: `{pull}`.
    """.format(
        push=click.style("wandb push weights.h5", bold=True),
        flags=click.style("import wandb; conf = wandb.sync(config=tf.__FLAGS__)", bold=True),
        config=click.style("wandb config set batch_size=10", bold=True),
        pull=click.style("wandb pull models/inception-v4", bold=True)
    ))

@cli.group()
@click.pass_context
@display_error
def config(ctx):
    """Manage this projects configuration.

Examples: 

    wandb config set param=2 --description="Some tunning parameter"
    wandb config del param                                          
    wandb config show
    """
    pass

@config.command("init", help="Initialize a directory with wandb configuration")
@display_error
def config_init(prompt=True):
    config_path = os.getcwd()+"/.wandb"
    config = Config()
    if os.path.isdir(config_path):
        if prompt:
            click.confirm(click.style("This directory is already initialized, should we overwrite it?", fg="red"), abort=True)
    else:
        os.mkdir(config_path)
    config.epochs_desc = "Number epochs to train over"
    config.epochs = 32
    config.persist()
    if prompt:
        click.echo("""Configuration initialized, use `wandb config set` to set parameters.  Then in your training script:

import wandb
conf = wandb.sync()
conf.batch_size
""")

@config.command(help="Show the current config")
@click.option("--format", help="The format to dump the config as", default="python", type=click.Choice(['python', 'yaml', 'json']))
@display_error
def show(format, changed=[], diff=False):
    if len(changed) == 0 and diff:
        click.secho("No parameters were changed", fg="red")
    elif diff:
        click.echo("%i parameters changed: " % len(changed))
    config = Config()
    if len(vars(config)) == 0:
        click.secho("No configuration found in this directory, run `wandb config init`", fg="red")
    if format == "yaml":
        click.echo("%s" % config)
    elif format == "json":
        click.echo(json.dumps(vars(config)))
    elif format == "python":
        res = ""
        for key in set(config.keys + changed):
            if config.desc(key):
                res += "# %s\n" % config.desc(key)
            style = None
            if key in changed:
                style = "green" if config.get(key) else "red"
            res += click.style("%s=%r\n" % (key, config.get(key)), bold=True if style is None else False, fg=style)
        click.echo(res)

@config.command("import", help="Import configuration parameters")
@click.option("--format", "-f", help="The format to parse the imported params", default="python", type=click.Choice(["python"]))
@click.pass_context
@display_error
def import_config(ctx, format):
    data = editor("# Paste python comments and variable definitions above")
    desc = None
    config = Config()
    imported = []
    if data:
        for line in data.split("\n"):
            if line.strip().startswith("#"):
                desc = line.strip(" #")
            elif "=" in line:
                try:
                    key, value = [str(part.strip()) for part in line.split("=")]
                    if len(value) == 0:
                        continue
                    config[key] = value
                    imported.append(key)
                    if desc:
                        config[key+"_desc"] = desc
                    desc = None
                except ValueError:
                    logging.error("Invalid line: %s" % line)
            else:
                logging.warn("Skipping line %s", line)
        config.persist()
    ctx.invoke(show, changed=imported, diff=True)

@config.command("set", help="Set config variables with key=value pairs")
@click.argument("key_values", nargs=-1)
@click.option("--description", "-d", help="A description for the config value if specifying one pair")
@click.pass_context
@display_error
def config_set(ctx, key_values, description=None):
    config = Config()
    if len(key_values) == 0:
        raise ClickException("Must specify at least 1 key value pair i.e. `wandb config set epochs=11`")
    if len(key_values) > 1 and description:
        raise ClickException("Description can only be specified with 1 key value pair.")
    changed = []
    for pair in key_values:
        try:
            key, value = pair.split("=")
        except ValueError:
            key = pair
            value = None
        if value:
            changed.append(key)
            config[str(key)] = value
        if description:
            config[str(key)+"_desc"] = description
    config.persist()
    ctx.invoke(show, changed=changed, diff=True)

@config.command("del", help="Delete config variables")
@click.argument("keys", nargs=-1)
@click.pass_context
@display_error
def delete(ctx, keys):
    config = Config()
    if len(keys) == 0:
        raise ClickException("Must specify at least 1 key i.e. `wandb config rm epochs`")
    changed = []
    for key in keys:
        del config[str(key)]
        changed.append(key)
    config.persist()
    ctx.invoke(show, changed=changed, diff=True)

if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-

__author__ = """Chris Van Pelt"""
__email__ = 'vanpelt@wandb.com'
__version__ = '0.4.18'

import types, sys, logging, os

__stage_dir__ = ".wandb/"
if not os.path.exists(__stage_dir__):
    __stage_dir__ = "/tmp/.wandb/"
    try:
        os.mkdir(__stage_dir__)
    except FileExistsError:
        pass

from .git_repo import GitRepo
from .api import Api, Error
from .sync import Sync
from .config import Config
from .results import Results
from .summary import Summary
from .history import History
from .keras import WandBKerasCallback

logging.basicConfig(
    filemode="w",
    filename=__stage_dir__+'debug.log',
    level=logging.DEBUG)

def push(*args, **kwargs):
    Api().push(*args, **kwargs)

def pull(*args, **kwargs):
    Api().pull(*args, **kwargs)

def sync(globs=['*'], **kwargs):
    api = Api()
    if api.api_key is None:
        raise Error("No API key found, run `wandb login` or set WANDB_API_KEY")
    #TODO: wandb describe
    sync = Sync(api, **kwargs)
    sync.watch(files=globs)
    return sync.config

__all__ = ["Api", "Error", "Config", "Results", "History", "Summary",
        "WandBKerasCallback"]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_streaming_log
----------------------------------

Tests for the `StreamingLog` module.
"""
import pytest
from freezegun import freeze_time
from wandb.streaming_log import StreamingLog
import time
from .api_mocks import upload_logs
import signal

@freeze_time("1981-12-09 12:00:01")
@pytest.fixture
def streamer():
    return StreamingLog("test")

@freeze_time("1981-12-09 12:00:01")
def test_basic(streamer):
    streamer.write("Foo bar baz\n")
    streamer.write("Some progress line\r")
    assert len(streamer.line_buffer.lines) == 2

@freeze_time("1981-12-09 12:00:01")
def test_dedupe_progress(streamer):
    streamer.write("[&...]\r")
    streamer.write("[&&..]\r")
    streamer.write("[&&&.]\r")
    streamer.write("[&&&&]\r")
    assert len(streamer.line_buffer.lines) == 1
    assert streamer.line_buffer.line_number == 1

@freeze_time("1981-12-09 12:00:02")
def test_push_by_length(request_mocker, upload_logs, streamer):
    mock = upload_logs(request_mocker, "test")
    streamer.write("1\n")
    streamer.write("2\n")
    streamer.write("3\n")
    streamer.write("4\n")
    streamer.write("5\n")
    assert len(streamer.line_buffer.lines) == 0
    assert mock.called

@freeze_time("1981-12-09 12:00:01")
def test_rate_limit(request_mocker, upload_logs, streamer):
    streamer.write("1\n")
    streamer.write("2\n")
    streamer.write("3\n")
    streamer.write("4\n")
    streamer.write("5\n")
    assert len(streamer.line_buffer.lines) == 5

@freeze_time("1981-12-09 12:00:01")
def test_partial_lines(request_mocker, upload_logs, streamer):
    streamer.write("bing ")
    streamer.write("bong\n")
    streamer.write("..")
    streamer.write("..\n")
    assert len(streamer.line_buffer.lines) == 2
    with open(streamer.tempfile.name, "r") as temp:
        assert temp.read() == """bing bong
....
"""

def test_raise_on_status(request_mocker, upload_logs, streamer):
    mock = upload_logs(request_mocker, "test", status_code=500)
    assert not mock.called

def test_push_by_stale(request_mocker, upload_logs, streamer):
    mock = upload_logs(request_mocker, "test")
    streamer.write("boom\n")
    assert len(streamer.line_buffer.lines) == 0
    assert mock.called

def test_heartbeat(request_mocker, upload_logs, streamer):
    mock = upload_logs(request_mocker, "test")
    #TODO: gotta be a better way to test this
    signal.alarm(1)
    time.sleep(1.1)
    assert mock.called